/**
 * ASM examples: examples showing how ASM can be use
 * Copyright (c) 2000-2011 INRIA, France Teleco
 * All rights reserved
 
 * Redistribution and use in source and binary forms, with or withou
 * modification, are permitted provided that the following condition
 * are met
 * 1. Redistributions of source code must retain the above copyrigh
 *    notice, this list of conditions and the following disclaimer
 * 2. Redistributions in binary form must reproduce the above copyrigh
 *    notice, this list of conditions and the following disclaimer in th
 *    documentation and/or other materials provided with the distribution
 * 3. Neither the name of the copyright holders nor the names of it
 *    contributors may be used to endorse or promote products derived fro
 *    this software without specific prior written permission
 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, TH
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOS
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS B
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, O
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT O
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINES
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER I
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED O
 * THE POSSIBILITY OF SUCH DAMAGE
 *
import java.io.FileOutputStream

import org.objectweb.asm.ClassWriter
import org.objectweb.asm.Label
import org.objectweb.asm.MethodVisitor
import org.objectweb.asm.Opcodes

/*
 * @author Eric Bruneto
 *
public class Compile extends ClassLoader 

    public static void main(final String[] args) throws Exception 
        // creates the expression tree corresponding t
        // exp(i) = i > 3 && 6 > 
        Exp exp = new And(new GT(new Var(0), new Cst(3)), new GT(new Cst(6)
                new Var(0)))
        // compiles this expression into an Expression clas
        Compile main = new Compile()
        byte[] b = exp.compile("Example")
        FileOutputStream fos = new FileOutputStream("Example.class")
        fos.write(b)
        fos.close()
        Class<?> expClass = main.defineClass("Example", b, 0, b.length)
        // instantiates this compiled expression class..
        Expression iexp = (Expression) expClass.newInstance()
        // ... and uses it to evaluate exp(0) to exp(9
        for (int i = 0; i < 10; ++i) 
            boolean val = iexp.eval(i, 0) == 1
            System.out.println(i + " > 3 && " + i + " < 6 = " + val)
        
    


/*
 * An abstract expression
 *
 * @author Eric Bruneto
 *
abstract class Exp implements Opcodes 

    /
     * Returns the byte code of an Expression class corresponding to thi
     * expression
     *
    byte[] compile(final String name) 
        // class heade
        ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS)
        cw.visit(V1_1, ACC_PUBLIC, name, null, "java/lang/Object"
                new String[] { Expression.class.getName() })

        // default public constructo
        MethodVisitor mv = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null
                null)
        mv.visitVarInsn(ALOAD, 0)
        mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V"
                false)
        mv.visitInsn(RETURN)
        mv.visitMaxs(1, 1)
        mv.visitEnd()

        // eval metho
        mv = cw.visitMethod(ACC_PUBLIC, "eval", "(II)I", null, null)
        compile(mv)
        mv.visitInsn(IRETURN)
        // max stack and max locals automatically compute
        mv.visitMaxs(0, 0)
        mv.visitEnd()

        return cw.toByteArray()
    

    /
     * Compile this expression. This method must append to the given code write
     * the byte code that evaluates and pushes on the stack the value of thi
     * expression
     *
    abstract void compile(MethodVisitor mv)


/*
 * A constant expression
 *
class Cst extends Exp 

    int value

    Cst(final int value) 
        this.value = value
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // pushes the constant's value onto the stac
        mv.visitLdcInsn(value)
    


/*
 * A variable reference expression
 *
class Var extends Exp 

    int index

    Var(final int index) 
        this.index = index + 1
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // pushes the 'index' local variable onto the stac
        mv.visitVarInsn(ILOAD, index)
    


/*
 * An abstract binary expression
 *
abstract class BinaryExp extends Exp 

    Exp e1

    Exp e2

    BinaryExp(final Exp e1, final Exp e2) 
        this.e1 = e1
        this.e2 = e2
    


/*
 * An addition expression
 *
class Add extends BinaryExp 

    Add(final Exp e1, final Exp e2) 
        super(e1, e2)
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // compiles e1, e2, and adds an instruction to add the two value
        e1.compile(mv)
        e2.compile(mv)
        mv.visitInsn(IADD)
    


/*
 * A multiplication expression
 *
class Mul extends BinaryExp 

    Mul(final Exp e1, final Exp e2) 
        super(e1, e2)
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // compiles e1, e2, and adds an instruction to multiply the two value
        e1.compile(mv)
        e2.compile(mv)
        mv.visitInsn(IMUL)
    


/*
 * A "greater than" expression
 *
class GT extends BinaryExp 

    GT(final Exp e1, final Exp e2) 
        super(e1, e2)
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // compiles e1, e2, and adds the instructions to compare the two value
        e1.compile(mv)
        e2.compile(mv)
        Label iftrue = new Label()
        Label end = new Label()
        mv.visitJumpInsn(IF_ICMPGT, iftrue)
        // case where e1 <= e2 : pushes false and jump to "end
        mv.visitLdcInsn(0)
        mv.visitJumpInsn(GOTO, end)
        // case where e1 > e2 : pushes tru
        mv.visitLabel(iftrue)
        mv.visitLdcInsn(1)
        mv.visitLabel(end)
    


/*
 * A logical "and" expression
 *
class And extends BinaryExp 

    And(final Exp e1, final Exp e2) 
        super(e1, e2)
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // compiles e
        e1.compile(mv)
        // tests if e1 is fals
        mv.visitInsn(DUP)
        Label end = new Label()
        mv.visitJumpInsn(IFEQ, end)
        // case where e1 is true : e1 && e2 is equal to e
        mv.visitInsn(POP)
        e2.compile(mv)
        // if e1 is false, e1 && e2 is equal to e1
        // we jump directly to this label, without evaluating e
        mv.visitLabel(end)
    


/*
 * A logical "or" expression
 *
class Or extends BinaryExp 

    Or(final Exp e1, final Exp e2) 
        super(e1, e2)
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // compiles e
        e1.compile(mv)
        // tests if e1 is tru
        mv.visitInsn(DUP)
        Label end = new Label()
        mv.visitJumpInsn(IFNE, end)
        // case where e1 is false : e1 || e2 is equal to e
        mv.visitInsn(POP)
        e2.compile(mv)
        // if e1 is true, e1 || e2 is equal to e1
        // we jump directly to this label, without evaluating e
        mv.visitLabel(end)
    


/*
 * A logical "not" expression
 *
class Not extends Exp 

    Exp e

    Not(final Exp e) 
        this.e = e
    

    @Overrid
    void compile(final MethodVisitor mv) 
        // computes !e1 by evaluating 1 - e
        mv.visitLdcInsn(1)
        e.compile(mv)
        mv.visitInsn(ISUB)
    

