/**
 * ASM examples: examples showing how ASM can be use
 * Copyright (c) 2000-2011 INRIA, France Teleco
 * All rights reserved
 
 * Redistribution and use in source and binary forms, with or withou
 * modification, are permitted provided that the following condition
 * are met
 * 1. Redistributions of source code must retain the above copyrigh
 *    notice, this list of conditions and the following disclaimer
 * 2. Redistributions in binary form must reproduce the above copyrigh
 *    notice, this list of conditions and the following disclaimer in th
 *    documentation and/or other materials provided with the distribution
 * 3. Neither the name of the copyright holders nor the names of it
 *    contributors may be used to endorse or promote products derived fro
 *    this software without specific prior written permission
 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, TH
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOS
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS B
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, O
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT O
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINES
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER I
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED O
 * THE POSSIBILITY OF SUCH DAMAGE
 *
package org.objectweb.asm.depend

import java.util.HashMap
import java.util.HashSet
import java.util.Map
import java.util.Set

import org.objectweb.asm.AnnotationVisitor
import org.objectweb.asm.ClassVisitor
import org.objectweb.asm.FieldVisitor
import org.objectweb.asm.Handle
import org.objectweb.asm.Label
import org.objectweb.asm.MethodVisitor
import org.objectweb.asm.Opcodes
import org.objectweb.asm.Type
import org.objectweb.asm.TypePath
import org.objectweb.asm.signature.SignatureReader
import org.objectweb.asm.signature.SignatureVisitor

/*
 * DependencyVisito
 *
 * @author Eugene Kulesho
 *
public class DependencyVisitor extends ClassVisitor 
    Set<String> packages = new HashSet<String>()

    Map<String, Map<String, Integer>> groups = new HashMap<String, Map<String, Integer>>()

    Map<String, Integer> current

    public Map<String, Map<String, Integer>> getGlobals() 
        return groups
    

    public Set<String> getPackages() 
        return packages
    

    public DependencyVisitor() 
        super(Opcodes.ASM5)
    

    // ClassVisito

    @Overrid
    public void visit(final int version, final int access, final String name
            final String signature, final String superName
            final String[] interfaces) 
        String p = getGroupKey(name)
        current = groups.get(p)
        if (current == null) 
            current = new HashMap<String, Integer>()
            groups.put(p, current)
        

        if (signature == null) 
            if (superName != null) 
                addInternalName(superName)
            
            addInternalNames(interfaces)
        } else 
            addSignature(signature)
        
    

    @Overrid
    public AnnotationVisitor visitAnnotation(final String desc
            final boolean visible) 
        addDesc(desc)
        return new AnnotationDependencyVisitor()
    

    @Overrid
    public AnnotationVisitor visitTypeAnnotation(final int typeRef
            final TypePath typePath, final String desc, final boolean visible) 
        addDesc(desc)
        return new AnnotationDependencyVisitor()
    

    @Overrid
    public FieldVisitor visitField(final int access, final String name
            final String desc, final String signature, final Object value) 
        if (signature == null) 
            addDesc(desc)
        } else 
            addTypeSignature(signature)
        
        if (value instanceof Type) 
            addType((Type) value)
        
        return new FieldDependencyVisitor()
    

    @Overrid
    public MethodVisitor visitMethod(final int access, final String name
            final String desc, final String signature, final String[] exceptions) 
        if (signature == null) 
            addMethodDesc(desc)
        } else 
            addSignature(signature)
        
        addInternalNames(exceptions)
        return new MethodDependencyVisitor()
    

    class AnnotationDependencyVisitor extends AnnotationVisitor 

        public AnnotationDependencyVisitor() 
            super(Opcodes.ASM5)
        

        @Overrid
        public void visit(final String name, final Object value) 
            if (value instanceof Type) 
                addType((Type) value)
            
        

        @Overrid
        public void visitEnum(final String name, final String desc
                final String value) 
            addDesc(desc)
        

        @Overrid
        public AnnotationVisitor visitAnnotation(final String name
                final String desc) 
            addDesc(desc)
            return this
        

        @Overrid
        public AnnotationVisitor visitArray(final String name) 
            return this
        
    

    class FieldDependencyVisitor extends FieldVisitor 

        public FieldDependencyVisitor() 
            super(Opcodes.ASM5)
        

        @Overrid
        public AnnotationVisitor visitAnnotation(String desc, boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public AnnotationVisitor visitTypeAnnotation(final int typeRef
                final TypePath typePath, final String desc
                final boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        
    

    class MethodDependencyVisitor extends MethodVisitor 

        public MethodDependencyVisitor() 
            super(Opcodes.ASM5)
        

        @Overrid
        public AnnotationVisitor visitAnnotationDefault() 
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public AnnotationVisitor visitAnnotation(final String desc
                final boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public AnnotationVisitor visitTypeAnnotation(final int typeRef
                final TypePath typePath, final String desc
                final boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public AnnotationVisitor visitParameterAnnotation(final int parameter
                final String desc, final boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public void visitTypeInsn(final int opcode, final String type) 
            addType(Type.getObjectType(type))
        

        @Overrid
        public void visitFieldInsn(final int opcode, final String owner
                final String name, final String desc) 
            addInternalName(owner)
            addDesc(desc)
        

        @Overrid
        public void visitMethodInsn(final int opcode, final String owner
                final String name, final String desc, final boolean itf) 
            addInternalName(owner)
            addMethodDesc(desc)
        

        @Overrid
        public void visitInvokeDynamicInsn(String name, String desc
                Handle bsm, Object... bsmArgs) 
            addMethodDesc(desc)
            addConstant(bsm)
            for (int i = 0; i < bsmArgs.length; i++) 
                addConstant(bsmArgs[i])
            
        

        @Overrid
        public void visitLdcInsn(final Object cst) 
            addConstant(cst)
        

        @Overrid
        public void visitMultiANewArrayInsn(final String desc, final int dims) 
            addDesc(desc)
        

        @Overrid
        public AnnotationVisitor visitInsnAnnotation(int typeRef
                TypePath typePath, String desc, boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public void visitLocalVariable(final String name, final String desc
                final String signature, final Label start, final Label end
                final int index) 
            addTypeSignature(signature)
        

        @Overrid
        public AnnotationVisitor visitLocalVariableAnnotation(int typeRef
                TypePath typePath, Label[] start, Label[] end, int[] index
                String desc, boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        

        @Overrid
        public void visitTryCatchBlock(final Label start, final Label end
                final Label handler, final String type) 
            if (type != null) 
                addInternalName(type)
            
        

        @Overrid
        public AnnotationVisitor visitTryCatchAnnotation(int typeRef
                TypePath typePath, String desc, boolean visible) 
            addDesc(desc)
            return new AnnotationDependencyVisitor()
        
    

    class SignatureDependencyVisitor extends SignatureVisitor 

        String signatureClassName

        public SignatureDependencyVisitor() 
            super(Opcodes.ASM5)
        

        @Overrid
        public void visitClassType(final String name) 
            signatureClassName = name
            addInternalName(name)
        

        @Overrid
        public void visitInnerClassType(final String name) 
            signatureClassName = signatureClassName + "$" + name
            addInternalName(signatureClassName)
        
    

    // --------------------------------------------

    private String getGroupKey(String name) 
        int n = name.lastIndexOf('/')
        if (n > -1) 
            name = name.substring(0, n)
        
        packages.add(name)
        return name
    

    private void addName(final String name) 
        if (name == null) 
            return
        
        String p = getGroupKey(name)
        if (current.containsKey(p)) 
            current.put(p, current.get(p) + 1)
        } else 
            current.put(p, 1)
        
    

    void addInternalName(final String name) 
        addType(Type.getObjectType(name))
    

    private void addInternalNames(final String[] names) 
        for (int i = 0; names != null && i < names.length; i++) 
            addInternalName(names[i])
        
    

    void addDesc(final String desc) 
        addType(Type.getType(desc))
    

    void addMethodDesc(final String desc) 
        addType(Type.getReturnType(desc))
        Type[] types = Type.getArgumentTypes(desc)
        for (int i = 0; i < types.length; i++) 
            addType(types[i])
        
    

    void addType(final Type t) 
        switch (t.getSort()) 
        case Type.ARRAY
            addType(t.getElementType())
            break
        case Type.OBJECT
            addName(t.getInternalName())
            break
        case Type.METHOD
            addMethodDesc(t.getDescriptor())
            break
        
    

    private void addSignature(final String signature) 
        if (signature != null) 
            new SignatureReader(signature
                    .accept(new SignatureDependencyVisitor())
        
    

    void addTypeSignature(final String signature) 
        if (signature != null) 
            new SignatureReader(signature
                    .acceptType(new SignatureDependencyVisitor())
        
    

    void addConstant(final Object cst) 
        if (cst instanceof Type) 
            addType((Type) cst)
        } else if (cst instanceof Handle) 
            Handle h = (Handle) cst
            addInternalName(h.getOwner())
            addMethodDesc(h.getDesc())
        
    

